from torch import nn, optim
import torch
from .fit import set_determenistic
import numpy as np

class mlp(nn.Module):
    def __init__(self, in_features, n_hidden, seed=None):
        
        set_determenistic(seed)
        super().__init__()
        
        
        self.in_features = in_features
        n_middle= int((in_features - n_hidden)/2) + n_hidden
        self.linear1 = nn.Linear(in_features=in_features, out_features=n_middle)
        self.linear2 = nn.Linear(in_features=n_middle, out_features=n_hidden) 
        self.linear3 = nn.Linear(in_features=n_hidden, out_features=n_middle) 
        self.linear4 = nn.Linear(in_features=n_middle, out_features=in_features) 
        self.Sigmoid = nn.Sigmoid()
    
    

    def forward(self, x):
        x = self.Sigmoid(self.linear1(x))
        x = self.Sigmoid(self.linear2(x))
        x = self.Sigmoid(self.linear3(x))
        y_pred = self.linear4(x)
        return y_pred

    def run_epoch(self, iterator, optimizer, criterion, points_ahead=1, phase='train', device=torch.device('cuda:0')):
        self.to(device)
        
        is_train = (phase == 'train')
        if is_train:
            self.train()
        else:
            self.eval()
        
        epoch_loss = 0


        all_y_preds = []
        with torch.set_grad_enabled(is_train):
            for i, (x,y) in enumerate(iterator):
                x,y = np.array(x),np.array(y) #df.index rif of
                
                x = torch.tensor(x).float().to(device).requires_grad_()
                y_true = torch.tensor(y).float().to(device)
                y_pred = self.forward(x)

                
                if phase == 'forecast':
                    all_y_preds.append(y_pred)
                    continue # in case of pahse = 'forecast' criterion is None
                        
                loss = criterion(y_pred,y_true)
                if is_train:
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()

                
                epoch_loss += loss.item()
        if phase != 'forecast':
            return epoch_loss / len(iterator)#, n_true_predicted / n_predicted
        else:
            return torch.cat(all_y_preds).detach().cpu().numpy()
            
            
            
            
            
            
class lstm(nn.Module):
    def __init__(self, in_features, n_hidden, seed=None):        
        super().__init__()
        set_determenistic(seed)
    
        n_middle= int((in_features - n_hidden)/2) + n_hidden
        
        self.in_features = in_features
        self.n_hidden = n_hidden
        self.n_middle = n_middle
        
        
        self.lstm1 = nn.LSTM(input_size=in_features,
                            hidden_size=n_middle,
                            batch_first =True)
        self.lstm2 = nn.LSTM(input_size=n_middle,
                            hidden_size=n_hidden,
                            batch_first =True)
        self.lstm3 = nn.LSTM(input_size=n_hidden,
                            hidden_size=n_middle,
                            batch_first =True)
        self.lstm4 = nn.LSTM(input_size=n_middle,
                            hidden_size=in_features,
                            batch_first =True)


        self.linear = nn.Linear(in_features=in_features, out_features=in_features)        
    
    
    def initHidden(self,batch_size,device):
        self.hidden_lstm1 = (
            torch.zeros(1, batch_size, self.n_middle).to(device),
            torch.zeros(1, batch_size, self.n_middle).to(device)
                            )
        self.hidden_lstm2 = (
            torch.zeros(1, batch_size, self.n_hidden).to(device),
            torch.zeros(1, batch_size, self.n_hidden).to(device)
                            )
        self.hidden_lstm3 = (
            torch.zeros(1, batch_size, self.n_middle).to(device),
            torch.zeros(1, batch_size, self.n_middle).to(device)
                            )
        self.hidden_lstm4 = (
            torch.zeros(1, batch_size, self.in_features).to(device),
            torch.zeros(1, batch_size, self.in_features).to(device)
                            )
    def forward(self, sequences):
        batch_size  = len(sequences)
        lstm_out1, self.hidden_lstm1 = self.lstm1(sequences, self.hidden_lstm1)
        lstm_out2, self.hidden_lstm2 = self.lstm2(lstm_out1, self.hidden_lstm2)
        lstm_out3, self.hidden_lstm3 = self.lstm3(lstm_out2, self.hidden_lstm3)
        lstm_out4, self.hidden_lstm4 = self.lstm4(lstm_out3, self.hidden_lstm4)
        
        # last_time_step = lstm_out4.reshape(-1, batch_size, self.in_features)[-1] # -1 is len_seq
        last_time_step = lstm_out4[:,-1,:]
        y_pred = self.linear(last_time_step)
        return y_pred

    def run_epoch(self, iterator, optimizer, criterion, phase='train', device=torch.device('cuda:0'), encod_decode_model=False, points_ahead=None):
        self.to(device)
        is_train = (phase == 'train')
        if is_train:
            self.train()
        else:
            self.eval()
        
        epoch_loss = 0
        all_y_preds = []
        with torch.set_grad_enabled(is_train):
            for i, (x,y) in enumerate(iterator):
                x,y = np.array(x),np.array(x)[:,-1,:] #!!! тут суть, см y
                self.initHidden(x.shape[0],device=device)
                
                x = torch.tensor(x).float().to(device).requires_grad_()
                y_true = torch.tensor(y).float().to(device)
                y_pred = self.forward(x)
                
                if phase == 'forecast':
                    all_y_preds.append(y_pred)
                    continue # in case of pahse = 'forecast' criterion is None
                        
                loss = criterion(y_pred,y_true)
                if is_train:
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()
                
                epoch_loss += loss.item()

        if phase != 'forecast':
            return epoch_loss / len(iterator)#, n_true_predicted / n_predicted
        else:
            return torch.cat(all_y_preds).detach().cpu().numpy()


"""
CDSDSDS
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec



from .univariate_funcs import confusion_matrix, single_average_delay, single_evaluate_nab
from .src import single_detecting_boundaries, check_errors, extract_cp_confusion_matrix


def evaluating(true, prediction,
               metric='nab',
               numenta_time=None,
               portion=None,
               anomaly_window_destenation='lefter',
               clear_anomalies_mode = True,
               intersection_mode='cut right window',
               table_of_coef=None,
               scale_func = "improved",
               scale_koef=1,               
               plot_figure=False,
               verbose=True
               ):
    """
    Parameters
    ----------
    true: variants:
        or: if one  dataset : pd.Series with binary int labels (1 is 
        anomaly, 0 is not anomaly);
        
        or: if one  dataset : list of pd.Timestamp of true labels, or [] 
        if haven't labels ;
        
        or: if one  dataset : list of list of t1,t2: left and right 
        detection, boundaries of pd.Timestamp or [[]] if haven't labels
        
        or: if many datasets: list (len of number of datasets) of pd.Series 
        with binary int labels;
        
        or: if many datasets: list of list of pd.Timestamp of true labels, or
        true = [ts,[]] if haven't labels for specific dataset;
        
        or: if many datasets: list of list of list of t1,t2: left and right 
        detection boundaries of pd.Timestamp;        
        If we haven't true labels for specific dataset then we must insert 
        empty list of labels: true = [[[]],[[t1,t2],[t1,t2]]]. 
        
        __True labels of anomalies or changepoints.
        It is important to have appropriate labels (CP or 
        anomaly) for corresponding metric (See later "metric")
        
    prediction: variants:
        or: if one  dataset : pd.Series with binary int labels
        (1 is anomaly, 0 is not anomaly);
        
        or: if many datasets: list (len of number of datasets) 
        of pd.Series with binary int labels.
        
        __Predicted labels of anomalies or changepoints.
        It is important to have appropriate labels (CP or 
        anomaly) for corresponding metric (See later "metric")
        
    metric: {'nab', 'binary', 'average_time', 'confusion_matrix'}. 
        Default='nab'
        Affects to output (see later: Returns)
        Changepoint problem: {'nab', 'average_time'}. 
        Standart AD problem: {'binary', 'confusion_matrix'}. 
        'nab' is Numenta Anomaly Benchmark metric
        
        'average_time' is both average delay or time to failure
        depend on situatuion.
        
        'binary': FAR, MAR, F1.
        
        'confusion_matrix' standart confusion_matrix for any point.
        
    numenta_time: 'str' for pd.Tiemdelta
        Width of detection window. Default=None.
        
    portion : float, default=0.1
        The portion is needed if numenta_time = None. 
        The width of the detection window in this case is equal 
        to a portion of the width of the length of prediction divided 
        by the number of real CPs in this dataset. Default=0.1.
        
    anomaly_window_destenation: {'lefter', 'righter', 'center'}. Defualt='right'
        The parameter of the location of the detection window relative to the anomaly. 
        'lefter'  : the detection window will be on the left side of the anomaly
        'righter' : the detection window will be on the right side of the anomaly
        'center'  : the scoring window will be positioned relative to the center of anom.
                  
    clear_anomalies_mode : boolean, default=True.
        True : then the `left value of a Scoring function is Atp and the 
        `right is Afp. Only the `first value inside the detection window is taken.
        False: then the `right value of a Scoring function is Atp and the 
        `left is Afp. Only the `last value inside the detection window is taken.

    intersection_mode: {'cut left widnow', 'cut right window', 'both'}. 
        Defualt='cut right window'
        The parameter will be used if the detection windows overlap for 
        true changepoints, which is generally undesirable and requires a 
        different approach than simply cropping the scoring window using 
        this parameter.
        'cut left window' : will cut the overlapping part of the left window
        'cut right window': will cut the intersecting part of the right window
        'both'            : will crop the intersecting portion of both the left 
        and right windows
    
    verbose:  booleant, default=True.
        If True, then output useful information
        
    plot_figure : booleant, default=False.
        If True, then drawing the score fuctions, detection windows and predictions
        It is used for example, for calibration the scale_koef. 

    table_of_coef (metric='nab'): pd.DataFrame of specific form. See bellow. 
        Application profiles of NAB metric.If Defaut is None:  
        table_of_coef = pd.DataFrame([[1.0,-0.11,1.0,-1.0],
                                      [1.0,-0.22,1.0,-1.0],
                                      [1.0,-0.11,1.0,-2.0]])
        table_of_coef.index = ['Standart','LowFP','LowFN']
        table_of_coef.index.name = "Metric"
        table_of_coef.columns = ['A_tp','A_fp','A_tn','A_fn']
        
    scale_func (metric='nab'): "default" of "improved". Defualt="improved".
        Scoring function in NAB metric.
        'default'  : standart NAB scoring function 
        'improved' : Our function for resolving disadvantages
        of standart NAB scoring function 
                  
    scale_koef : float > 0. Defualt=1.0. 
        Smoothing factor. The smaller it is, 
        the smoother the scoring function is.
                  
    Returns
    ----------
    metrics : value of metrics, depend on metric 
        'nab': tuple
            - Standart profile, float
            - Low FP profile, float
            - Low FN profilet
        'average_time': tuple
            - Average time (average delay, or time to failure)
            - Missing changepoints, int
            - FPs, int
            - Number of true changepoints, int
        'binary': tuple
            - F1 metric, float
            - False alarm rate, %, float
            - Missing Alarm Rate, %, float
        'binary': tuple
            - TPs, int
            - TNs, int
            - FPs, int
            - FNS, int 
                        
    """
    # ююююююююююююююююююююююююююююююююююююююююююююююююю
    ## проверки 
    
    assert isinstance(true, pd.Series) or isinstance(true, list)
    # проверки prediction
    if isinstance(prediction, pd.Series):
        true = [true]
        prediction = [prediction]
    elif isinstance(prediction,list):
        if  not all(isinstance(my_el, pd.Series) for my_el in prediction):
            raise Exception('Uncorrect format for prediction')
    else:
        raise Exception('Uncorrect format for prediction')
        
    # проверки длин датасетов: Number of dataset unequal
    assert len(true) == len(prediction)
    
   
    
    # проверки true
    input_variant= check_errors(true)
    
    
    def check_sort(my_list,input_variant):
        for dataset in my_list:
            if input_variant==2:
                assert all(np.sort(dataset)==np.array(dataset))
            elif input_variant==3:
                assert all(np.sort(np.concatenate(dataset))==np.concatenate(dataset))
            elif input_variant==1:
                assert all(dataset.index.values == dataset.sort_index().index.values)
    check_sort(true,input_variant)
    check_sort(prediction,1)
    
    # ююююююююююююююююююююююююююююююююююююююююююююююююю
    # part 2. To detected boundaries 
    
    if (portion is None) and (numenta_time is None) and (input_variant!=3):
        portion= 0.1
        print( f'Since you not choose numenta_time and portion, then portion will be {portion}')      
        
    
   
    if input_variant==1:
        detecting_boundaries = [single_detecting_boundaries(true_series = true[i], 
                                                            true_list_ts = None,
                                                            prediction = prediction[i],
                                                            numenta_time=numenta_time,
                                                            portion=portion,
                                                            anomaly_window_destenation=anomaly_window_destenation,
                                                            intersection_mode=intersection_mode)
                                for i in range(len(true))]
          
    elif input_variant==2:
        detecting_boundaries = [single_detecting_boundaries(true_series = None,
                                                            true_list_ts=true[i], 
                                                            prediction=prediction[i], 
                                                            numenta_time=numenta_time,
                                                            portion=portion,
                                                            anomaly_window_destenation=anomaly_window_destenation,
                                                            intersection_mode=intersection_mode)
                                    for i in range(len(true))]
          
    elif input_variant==3:
        detecting_boundaries = true.copy()
        # NExt anti fool system [[[t1,t2]],[]] -> [[[t1,t2]],[[]]] 
        for i in range(len(detecting_boundaries)): 
            if len(detecting_boundaries[i])==0:
                    detecting_boundaries[i]=[[]]   
    else:
        raise Exception('Unknown format for true')
        
        


    
    # ююююююююююююююююююююююююююююююююююююююююююююююююю
    # part 3. To compute metric
    
    #print(detecting_boundaries)
    if plot_figure:
        num_datasets = len(true)
        if  ((metric=='binary') or (metric=='confusion_matrix')) \
            and (input_variant==1):
            f = plt.figure(figsize=(16,5*num_datasets))
            grid = gridspec.GridSpec(num_datasets, 1)
            for i in range(num_datasets):
                globals()['ax'+str(i)] = f.add_subplot(grid[i])
                prediction[i].plot(ax=globals()['ax'+str(i)],label='pred',marker='o')
                true[i].plot(ax=globals()['ax'+str(i)],label='true',marker='o')
                globals()['ax'+str(i)].legend()
            plt.show()
        else:
            from .univariate_funcs import my_scale
            f = plt.figure(figsize=(16,5*num_datasets))
            grid = gridspec.GridSpec(num_datasets, 1)
            detalization = 100
            for i in range(num_datasets):
                globals()['ax'+str(i)] = f.add_subplot(grid[i])
                print_legend_boundary=True
                def plot_cp(couple,anomaly_window_destenation,ax,label):
                    if anomaly_window_destenation=='lefter':
                        ax.axvline(couple[1],c='r',label=label)
                    elif anomaly_window_destenation=='righter':
                        ax.axvline(couple[0],c='r',label=label)
                    elif anomaly_window_destenation=='center':
                        ax.axvline(couple[0]+((couple[1]-couple[0])/2),c='r',label=label)
                
                for couple in detecting_boundaries[i]:
                    if len(couple)>0:
                        globals()['ax'+str(i)].axvspan(couple[0],couple[1], alpha=0.5, color='green',
                        label='detection \nboundary' if print_legend_boundary else None)
                        nab = pd.Series(my_scale(plot_figure=True,detalization=detalization),
                                        index=pd.date_range(couple[0],couple[1],periods=detalization))
                        nab.plot(ax=globals()['ax'+str(i)], linewidth=0.4, color='brown',
                        label='nab scoring func' if print_legend_boundary else None)
                        plot_cp(couple,anomaly_window_destenation,globals()['ax'+str(i)],
                                label='Changepoint' if print_legend_boundary else None)
                        print_legend_boundary = False                                                
                    else: 
                        pass 
                prediction[i].plot(ax=globals()['ax'+str(i)],label='pred', marker='o')
                globals()['ax'+str(i)].legend()
            plt.show()
        
            
            
    
    
    if  metric=='nab':
        matrix = np.zeros((3,3))
        for i in range(len(prediction)):
            matrix_ = single_evaluate_nab(detecting_boundaries[i],
                                          prediction[i],
                                          table_of_coef=table_of_coef,
                                          clear_anomalies_mode = clear_anomalies_mode,
                                          scale_func = scale_func,
                                          scale_koef=scale_koef,
                                          plot_figure=plot_figure)
            matrix = matrix + matrix_      
                    
        results = {}
        desc = ['Standart', 'LowFP', 'LowFN'] 
        for t, profile_name in enumerate(desc):
            results[profile_name] = round(100*(matrix[0,t]-matrix[1,t])/(matrix[2,t]-matrix[1,t]), 2)
            if verbose:
                print(profile_name,' - ', results[profile_name])
        return results
    
    elif metric=='average_time':
        missing, detectHistory, FP, all_true_anom = 0, [], 0, 0
        for i in range(len(prediction)):
            missing_, detectHistory_, FP_, all_true_anom_ = single_average_delay(detecting_boundaries[i], 
                                                                                 prediction[i],
                                                                                 anomaly_window_destenation=anomaly_window_destenation,
                                                                                 clear_anomalies_mode=clear_anomalies_mode)
            missing, detectHistory, FP, all_true_anom = missing+missing_, detectHistory+detectHistory_, FP+FP_, all_true_anom+all_true_anom_
        add = np.mean(detectHistory)
        if verbose:
            print('Amount of true anomalies',all_true_anom)
            print(f'A number of missed CPs = {missing}')
            print(f'A number of FPs = {int(FP)}')
            print('Average time', add)
        return add, missing, int(FP), all_true_anom
    
    elif (metric=='binary') or (metric=='confusion_matrix') :
        
        if all(isinstance(my_el, pd.Series) for my_el in true):
            TP,TN,FP,FN = 0,0,0,0
            for i in range(len(prediction)):
                TP_,TN_,FP_,FN_ = confusion_matrix(true[i],prediction[i])
                TP,TN,FP,FN = TP+TP_,TN+TN_,FP+FP_,FN+FN_       
        else:
            print('For this metric it is better if you use pd.Series format for true \nwith common index of true and prediciton')
            TP,TN,FP,FN = 0,0,0,0
            for i in range(len(prediction)):
                dict_cp_confusion = extract_cp_confusion_matrix(detecting_boundaries[i],prediction[i],binary=True)
                TP+=np.sum([len(dict_cp_confusion['TPs'][window][1]) for window in dict_cp_confusion['TPs']])
                FP+=len(dict_cp_confusion['FPs'])
                FN+=len(dict_cp_confusion['FNs'])
                TN+= len(prediction[i]) - TP - FP - FN
            
            
        if metric=='binary':
            f1 = round(TP/(TP+(FN+FP)/2), 2)
            far = round(FP/(FP+TN)*100,2)
            mar = round(FN/(FN+TP)*100,2)
            if verbose:
                print(f'False Alarm Rate {far} %' )
                print(f'Missing Alarm Rate {mar} %')
                print(f'F1 metric {f1}')
            return f1,far,mar
        
        elif metric=='confusion_matrix':
            if verbose:
                print('TP',TP)
                print('TN',TN)
                print('FP',FP)
                print('FN',FN)
            return TP,TN,FP,FN 
    else:
        raise Exception("Choose the perfomance metric")



def import_module_from_GitHub(url,
                    saving_path = './Downloads/'):
    """ 
    A function that imports a Python module published on the Internet.

    Parameters
    ----------
    url : str
        Link to a module. Can be GitHub link to module
        or raw.githubusercontent.com or other, which 
        return HTTP get correctly.
    
    saving_path : str
        Path to saving model. 

    Returns
    -------
    module : module
        Then you can use this module
        Result as: import url 
        
    Examples
    --------
    >>> import_from_GitHub(https://github.com/Gabriel-p/minenergy/blob/master/minenergy.py)
       

    Notes
    -------
    1 : Синтаксис
        Хотелось бы from url import func1 чтобы работало. 
    
    """
    
    
    host = url.split('/')[2]
    if host == 'raw.githubusercontent.com':
        pass
    elif host == 'github.com':
        url = url.replace('github.com','raw.githubusercontent.com').replace('blob/','')
    
    local_filename = url.split('/')[-1]
    assert local_filename.split('.')[-1] == 'py'
    
    # === saving file  == = 
    # NOTE the stream=True parameter below
    import requests
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(saving_path + local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192): 
                f.write(chunk)
    
    # === import file  == =
    import importlib.util
    spec = importlib.util.spec_from_file_location(local_filename, saving_path + local_filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module
    
    
    
# repo 

# import git

# repo = git.Repo.clone_from(
#     'https://github.com/gitpython-developers/GitPython',
#     './git-python')   
    
    
    



from torch import nn, optim
import torch
from .fit import set_determenistic
import numpy as np



class MLP(nn.Module):
    def __init__(self, tuple_layers, dropout=0.5, seed=None):
        set_determenistic(seed)
        super(MLP, self).__init__()
        
        if len(tuple_layers)<3:
            raise
        self.seq = nn.Sequential()
        layers = (nn.Linear(in_features=tuple_layers[i], 
                           out_features=tuple_layers[i+1])  for i in range(len(tuple_layers)-2))
        drops = (nn.Dropout(p=dropout) for i in range(len(tuple_layers)-2))
        sigmoids = ( nn.Sigmoid()      for i in range(len(tuple_layers)-2))
        seq = (j for sub in tuple(zip(layers,sigmoids,drops)) for j in sub)
        self.seq = nn.Sequential(*seq,
                                  nn.Linear(in_features=tuple_layers[-2], 
                                            out_features=tuple_layers[-1]))
    def forward(self, x):
        y  = self.seq(x)
        return y

    def run_epoch(self, iterator, optimizer, criterion, points_ahead=1, phase='train', 
                  device=torch.device('cuda:0'), encod_decode_model=False):
        self.to(device)
        
        is_train = (phase == 'train')
        if is_train:
            self.train()
        else:
            self.eval()
        
        epoch_loss = 0

        all_y_preds = []
        with torch.set_grad_enabled(is_train):
            for i, (x,y) in enumerate(iterator):
                x,y = np.array(x),np.array(y) #df.index rif of
    
                x = torch.tensor(x).float().to(device).requires_grad_()
                y_true = torch.tensor(y).float().to(device)
                y_pred = self.forward(x)
                if phase == 'forecast':
                    all_y_preds.append(y_pred)
                    continue # in case of pahse = 'forecast' criterion is None

                        
                loss = criterion(y_pred,y_true)
                if is_train:
                  optimizer.zero_grad()
                  loss.backward()
                  optimizer.step()

                
                epoch_loss += loss.item()
        if phase != 'forecast':
            return epoch_loss / len(iterator)#, n_true_predicted / n_predicted
        else:
            return torch.cat(all_y_preds).detach().cpu().numpy()


import pandas as pd 

def value_counts_interval(array,itervals):
    """
    input : np.array, list of values
    retrun : pd.series
    """
    names = [f"до {itervals[0]}"]
    quantity = [len(array[array < itervals[0]])]
    for i in range(len(itervals)-1):
        quantity.append(len(array[(array >= itervals[i]) & (array < itervals[i+1])]))
        names.append(f'c {itervals[i]} до {itervals[i+1]}')
    names += [f"от {itervals[-1]}"]
    quantity += [len(array[array >= itervals[-1]])]
    ts = pd.Series(quantity,index=names)    
    return ts


class R_to_I():
    """
    A class that allows you to make a binary multidimensional time series from a 
    multidimensional time series of Real values. It is needed to adapt the 
    input for spike neural networks. API uses skirit-learn templates
    
    [Ru]: Класс который позволяет из многомерного временного ряда реальных значений
    сделать бинарный многомерный врменной ряд. Нужен для адаптации входных для спайковых 
    нейронных сетей. Для API используются шаблоны skirit-learn
    """
    
    def koefI_to_koefSeqI(self,ts):
        """
        A function that converts the number 5 to 5 consecutive impulses in frames
        
        Parameters
        ----------
        ts : pd.Series 
            The initial series with Real numbers
        
        Returns
        -------
        new_ts : pd.Series 
            The output series with binary labels
        
        Функция которая преобрарузет число 5 в 5 последовательных импульсов 
        """
        buffer = 0
        new_array = []
        for ind,val in ts.iteritems():
            if val !=0:
                buffer += val 
            new_val = 1 if buffer > 0 else 0
            new_array.append(new_val)
            buffer = buffer - 1 if buffer > 0 else 0
        new_ts = pd.Series(new_array,index=ts.index)
        return new_ts
    
    
    def __init__(self,verbose=False):
        """
        A class that allows you to make a binary multidimensional time series from a 
        multidimensional time series of Real values. It is needed to adapt the 
        input for spike neural networks. API uses skirit-learn templates. Useful if the signal 
        is being written by aperture.

        [Ru]: Класс который позволяет из многомерного временного ряда реальных значений
        сделать бинарный многомерный врменной ряд. Нужен для адаптации входных для спайковых 
        нейронных сетей. Для API используются шаблоны skirit-learn. Полезен если сигнал пишется
        по апертуре. 
        
        Parameters
        ----------
        verbose : boolean, default = False
            If True, additional infomation will be provided
        """
        self.verbose = verbose
    
    def _check_input_(self,x):
        assert any(np.isnan(x))
        assert len(x)!=0
        x.columns = x.columns.astype(str)
        
    def fit(self,x):
        """
        Fit the model with x

        [Ru]: Выделение апертуры в соотвествии с х
        
        Parameters
        ----------
        x : pd.DataFrame
            Training data. 
        """
        self._check_input_(x)      
        df = x.dropna().diff().dropna().abs()
        vector_koef = []
        self.block_columns = [] #колонки, которые не меняются вообще, то есть константы, их не будет после 
                                #трансформации, а после inverse_transform они будут константой той же
        for col in df:
            ts = df[col]
            ts = ts[ts!=0.]
            if len(ts)!=0:
                vector_koef.append(ts.value_counts().index[0])
            else: 
                self.block_columns.append(col)
        self.all_columns = x.columns
        self.norm_columns =  self.all_columns[~self.all_columns.isin(self.block_columns)]
        self.vector_koef = pd.DataFrame(np.array(vector_koef).reshape(1,-1), 
                                        columns = self.norm_columns)
        return self
    
    def transform(self,x):
        """
        Converting real values x to binary.

        [Ru]: Преобразование реальных значений в бианрные. 
        
        Parameters
        ----------
        x : pd.DataFrame
            Data which we want to transform from real values to binary values. 
        """
        self._check_input_(x)  
        result= x.copy()
        self.vector_0 = result.iloc[0:1] 
        self.index = result.index   
        result = result[self.norm_columns]
        result = (result.diff() / self.vector_koef.values).dropna().round().astype(int)        
        print('Fistt stage done')
        # разделяем на положительные и отрицательные признаки, посути увеличив количесто признаков в 2 раза 
        result = pd.concat([result[result>=0].add_suffix('_pos'),
                            result[result<=0].add_suffix('_neg')] ,
                            axis=1).fillna(0).astype(int).abs()
        print('Second stage done')
        if self.verbose:
            print('Максимальные величины импульсов (не единичные еще) для признаков')
            display_df(result.max())
            print('Распределение величин импульсов по всей таблице')
            display_df(pd.Series(result.values.ravel()).value_counts())
        print('Thied stage done')
        #  Функция которая преобрарузет число 5 в 5 последовательных импульсов 
        result = result.apply(self.koefI_to_koefSeqI)
        print('thourh stage done')
        return result # первый нан
    
    def fit_trasform(self,x):
        """
        Fitting and converting real values x to binary.

        [Ru]: Преобразование реальных значений в бианрные. 
        
        Parameters
        ----------
        x : pd.DataFrame
            Data which we want to transform from real values to binary values. 
        """
        
        return self.fit(x).transform(x)
    
    def inverse_tranform(self,x):
        """
        Fitting and converting real values x to binary.

        [Ru]: Преобразование реальных значений в бианрные. 
        
        Parameters
        ----------
        x : pd.DataFrame
            Data which we want to transform from binary values to real values.
            
        Notes
        ----------
        Очень важно иметь общие индекс с transform(self,x)
        """        
        self._check_input_(x)
        result = x.copy()
        neg = result.filter(regex='_neg',axis=1)
        neg.columns = neg.columns.str.strip('_neg')
        result = result.filter(regex='_pos',axis=1).values - neg
        del neg
        print('five stage done')
        result = result *  self.vector_koef.values.astype(int)
        result = pd.concat([self.vector_0,result])
        assert (result.index == self.index).all()
        result = result.cumsum()
        
        for col in self.block_columns:
            result[col] = self.vector_0[col].values[0]
        result = result[self.all_columns]
        return result


"""
Very useful librray
"""
import numpy as np

class MeshLoader:
    def __init__(self, params):
        """Only list in list, without оборачиваетелй numpy
		# mesh_alpha = np.arange(10)
		# mesh_beta = np.arange(10)
		# for k,z in enumerate(MeshLoader([[0.1,0.2,0.3],['a','b']])):
		#     print(z)
		"""
        self.params = params
        _my_array = np.ones(tuple(len(obj) for obj in params))
        self.indexes = []
        for index, values in np.ndenumerate(_my_array):
            self.indexes.append(index)

    def __iter__(self):
        self.i = -1
        return self

    def __next__(self):
        if self.i < len(self.indexes)-1:
            self.i+=1
            return tuple(self.params[from_parametr][self.indexes[self.i][from_parametr]] for from_parametr in range(len(self.params)))
        else:
            raise StopIteration

class Loader:
    def __init__(self, X,y, batch_size,shuffle=True, random_state = None):
        """
        X, y  are lists
        """
        if shuffle==True:
            np.random.seed(random_state)
            indices = np.array(range(len(X)))
            np.random.shuffle(indices)
            self.X = [X[ind] for ind in indices]
            self.y = [y[ind] for ind in indices]
        else:
            self.X = X
            self.y = y
        self.batch_size = batch_size

    def __iter__(self):
        self.i = - self.batch_size
        return self

    def __next__(self):
        if self.i+self.batch_size < len(self.X):
            self.i+=self.batch_size
            return self.X[self.i:self.i+self.batch_size], self.y[self.i:self.i+self.batch_size]
        elif self.i+2*self.batch_size < len(self.X):        
            return self.X[self.i:], self.y[self.i:]
        else:
            raise StopIteration
            
    def __len__(self):
        return len(np.arange(0,len(self.X),self.batch_size))
		        


import pandas as pd 
import matplotlib.pyplot as plt
import numpy as np

from .src import extract_cp_confusion_matrix, filter_detecting_boundaries

def confusion_matrix(true,prediction):
    true_ = true == 1 
    prediction_ = prediction == 1
    TP = (true_ & prediction_).sum()
    TN = (~true_ & ~prediction_).sum()
    FP = (~true_ & prediction_).sum()
    FN = (true_ & ~prediction_).sum()
    return TP,TN,FP,FN
    
def single_average_delay(detecting_boundaries, prediction,anomaly_window_destenation,clear_anomalies_mode):
    """
    anomaly_window_destenation: 'lefter', 'righter', 'center'. Defualt='right'
    """    
    detecting_boundaries = filter_detecting_boundaries(detecting_boundaries)  
    point = 0 if clear_anomalies_mode else -1
    dict_cp_confusion = extract_cp_confusion_matrix(detecting_boundaries, prediction,point=point)
    
    missing = 0
    detectHistory = []
    all_true_anom = 0
    FP = 0
    
    FP += len(dict_cp_confusion['FPs'])
    missing += len(dict_cp_confusion['FNs'])
    all_true_anom+=len(dict_cp_confusion['TPs']) + len(dict_cp_confusion['FNs']) 
    
    if anomaly_window_destenation=='lefter':
        def average_time(output_cp_cm_tp):
            return output_cp_cm_tp[2]-output_cp_cm_tp[1]
    elif anomaly_window_destenation=='righter':
        def average_time(output_cp_cm_tp):
            return output_cp_cm_tp[1]-output_cp_cm_tp[0]
    elif anomaly_window_destenation=='center':
         def average_time(output_cp_cm_tp):
            return output_cp_cm_tp[1]-(output_cp_cm_tp[0]+ (output_cp_cm_tp[2]- output_cp_cm_tp[0])/2)  
    else:
        raise Exception("Choose anomaly_window_destenation")
    
    for fp_case_window in dict_cp_confusion['TPs']:
        detectHistory.append(average_time(dict_cp_confusion['TPs'][fp_case_window])) 
    return missing, detectHistory, FP, all_true_anom



def my_scale(fp_case_window=None,A_tp=1,A_fp=0,koef=1,detalization=1000,clear_anomalies_mode=True,plot_figure=False):
    """ts - участок на котором надо жахнуть окно """
    x = np.linspace(-np.pi/2,np.pi/2,detalization)
    x = x if clear_anomalies_mode else x[::-1]
    y = (A_tp-A_fp)/2*-1*np.tanh(koef*x)/(np.tanh(np.pi*koef/2)) + (A_tp-A_fp)/2 + A_fp
    if not plot_figure:
        event =int( (fp_case_window[1] - fp_case_window[0]) / \
            (fp_case_window[-1] - fp_case_window[0])*detalization)
        if event>= len(x):
            event = len(x) -1 
        score = y[event]
        return score
    else:
        return y    
    
def single_evaluate_nab(detecting_boundaries,
                        prediction, 
                        table_of_coef=None,
                        clear_anomalies_mode = True,
                        scale_func = "improved",
                        scale_koef=1,
                        plot_figure=True #TODO
                       ):
    """
    
    detecting_boundaries: list of list of two float values
                The list of lists of left and right boundary indices
                for scoring results of labeling if empty. Can be [[]], or [[],[t1,t2],[]]                
    table_of_coef: pandas array (3x4) of float values
                Table of coefficients for NAB score function
                indeces: 'Standart','LowFP','LowFN'
                columns:'A_tp','A_fp','A_tn','A_fn'
                
    scale_func {default}, improved
    недостаттки scale_func default  -
    1 - зависит от относительного шага, а это значит, что если 
    слишком много точек в scoring window то перепад будет слишком
    жестким в середение. 
    2-   то самая левая точка не равно  Atp, а права не равна Afp
    (особенно если пррименять расплывающую множитель)

    clear_anomalies_mode тогда слева от границы Atp срправа Afp,
    иначе fault mode, когда слева от границы Afp срправа Atp
    """

#     def sigm_scale(len_ts, A_tp, A_fp, koef=1):
#         x = np.arange(-int(len_ts/2), len_ts - int(len_ts/2))
        
#         x = x if clear_anomalies_mode else x[::-1]
#         y = (A_tp-A_fp)*(1/(1+np.exp(5*x*koef))) + A_fp
#         return y
#     def my_scale(len_ts,A_tp,A_fp,koef=1):
#         """ts - участок на котором надо жахнуть окно """
#         x = np.linspace(-np.pi/2,np.pi/2,len_ts)
#         x = x if clear_anomalies_mode else x[::-1]
#         # Приведение если неравномерный шаг.
#         #x_new = x_old * ( np.pi / (x_old[-1]-x_old[0])) - x_old[0]*( np.pi / (x_old[-1]-x_old[0])) - np.pi/2
#         y = (A_tp-A_fp)/2*-1*np.tanh(koef*x)/(np.tanh(np.pi*koef/2)) + (A_tp-A_fp)/2 + A_fp
#         return y 

    if scale_func == "improved":
        scale_func = my_scale
#     elif scale_func == "default":
#         scale_func = sigm_scale
    else:
        raise Exception("choose the scale_func")
    
    #filter
    detecting_boundaries = filter_detecting_boundaries(detecting_boundaries)  
    
    if table_of_coef is None:
        table_of_coef = pd.DataFrame([[1.0,-0.11,1.0,-1.0],
                             [1.0,-0.22,1.0,-1.0],
                              [1.0,-0.11,1.0,-2.0]])
        table_of_coef.index = ['Standart','LowFP','LowFN']
        table_of_coef.index.name = "Metric"
        table_of_coef.columns = ['A_tp','A_fp','A_tn','A_fn']

    #GO
    point = 0 if clear_anomalies_mode else -1
    dict_cp_confusion = extract_cp_confusion_matrix(detecting_boundaries, prediction,point=point)

    Scores, Scores_perfect, Scores_null=[], [], []
    for profile in ['Standart', 'LowFP', 'LowFN']:       
        A_tp = table_of_coef['A_tp'][profile]
        A_fp = table_of_coef['A_fp'][profile]
        A_fn = table_of_coef['A_fn'][profile]
        
        score = 0
        score += A_fp*len(dict_cp_confusion['FPs'])
        score += A_fn*len(dict_cp_confusion['FNs'])
        for fp_case_window in dict_cp_confusion['TPs']:
            set_times = dict_cp_confusion['TPs'][fp_case_window]
            score+=scale_func(set_times,A_tp,A_fp,koef=scale_koef)

        Scores.append(score)
        Scores_perfect.append(len(detecting_boundaries)*A_tp)
        Scores_null.append(len(detecting_boundaries)*A_fn)
        
    return np.array([np.array(Scores),np.array(Scores_null), np.array(Scores_perfect)])


# косяк, мой подсчет метрик не работает если там нет трушных 1
"""
CDSDSDS
"""

from sklearn.model_selection import train_test_split
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import torch
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

 
def ts_train_test_split(df, len_seq, 
                     points_ahead=1, gap=0, shag=1, intersection=True,
                     test_size=None,train_size=None, random_state=None, what_to_shuffle='train'):
    """
    A function that splits the time series into train and test subsets  

    Parameters
    ----------
    df : pd.DataFrame
        Array of shape (n_samples, n_features) with pd.timestamp index.
    
    points_ahead : int, default=0
        How many points ahead we predict, reflected in y
         
    gap :  int, default=0
        How many points between train and test. Relatively speaking, 
        if the "right" point of train is t, then the first point of the test 
        is t + gap +1. The parameter is designed to be able to predict 
        one point after a large additional time interval.
    
    shag :  int, default=1.
        Sample generation step. If the first point was t for 
        the 1st sample of the train, then for the 2nd sample of the train 
        it will be t + shag if intersection=True, otherwise the same 
        but without intersections of the series values.

    intersection :  bool, default=True
       The presence of series values (of one point in time) in different 
       samples for the train set and and separately for the test test. 
       The train and the test never have common points.
    
    test_size : float or int or timestamp for df, or list of timestamps, default=0.25. 
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. 
        If int, represents the absolute number of test samples. If None, the value is set to the
        complement of the train size. If ``train_size`` is also None, it will
        be set to 0.25. *
        If timestamp for df, for X_test we will use set from df[t:] 
        If list of timestamps [t1,t2], for X_test we will use set from df[t1:t2] 
        Can be 0, then it will return the X,y values in X_train, y_train. 
        
    train_size : float or int, default=None
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size. 
        If timestamp for df, for X_train we will use set for train from df[:t] 
        If list of timestamps [t1,t2], for X_train we will use set for train from df[t1:t2] 
        Can be 0, then it will return the X,y values in X_test, y_test.
        
    what_to_shuffle: {'nothing', 'all','train'}, str. Default = 'train'. 
        In the case of 'train' we random shuffle only X_train, and y_train. 
        Test samples are unused for the shuffle. Any sample from X_test is later 
        than any sample from X_train. This is also true for respectively
        In case of 'all' in analogy with sklearn.model_selection.train_test_split
        In case of 'nothing' shuffle is not performed.
        
    random_state : int, RandomState instance or None, default=None
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.*  

    Returns
    -------
    (X_train, X_test, y_train, y_test) : tuple 
        Tuple containing train-test split of inputs
    
    
    Examples
    --------
    >>> X = np.ones((4, 3))
    >>> y = np.ones(4)
    >>> sklearn_template(X, y)
    (z, xmin, xmax)  # this should match the actual output
    """
    
    
      # проблема вклюения правых границ
        
    
    
    # Default settings
    if (train_size is None) and (test_size is None):
        train_size = 0.75
        test_size  = 0.25

    assert len_seq + points_ahead + gap  <= len(df)
    assert all(np.sort(df.index)==np.array(df.index))



    ####### Part I: Вычисление подряд возможных samples
    x_start=0
    x_end= x_start + len_seq
    y_start = x_end + gap 
    y_end = y_start + points_ahead
    if intersection:
        # ради вычислительной нагрузки такой кастыль
        def compute_new_x_start(x_start,y_end,shag):
            return x_start + shag
    else:
        def compute_new_x_start(x_start,y_end,shag):
            return y_end + shag -1
    X = []
    y = []
    while y_end <= len(df):
        X.append(df[x_start:x_end])
        y.append(df[y_start:y_end])

        x_start= compute_new_x_start(x_start,y_end,shag)
        x_end= x_start + len_seq
        y_start = x_end + gap
        y_end = y_start + points_ahead

    ####### Part 2: Выход на train_sample_numbers и test_sample_numbers 

    train_sample_numbers = None 
    test_sample_numbers  = None
    if isinstance(train_size,(pd.Timestamp,list)) or isinstance(test_size,(pd.Timestamp,list)):
        df_vot = pd.DataFrame({'start':[_df.index[0] for _df in X], 'end': [_df.index[-1] for _df in y]}).reset_index()  
        def check_list_assert(my_list):
            for val in my_list:
                assert isinstance(val, pd.Timestamp)
        if isinstance(train_size,(pd.Timestamp,list)):
            if isinstance(train_size,pd.Timestamp):
                t_train_left  = None
                t_train_right = train_size
            elif isinstance(train_size,list):
                check_list_assert(train_size)
                t_train_left = train_size[0]
                t_train_right = train_size[1]
            train_indexes = df_vot.set_index('end').truncate(t_train_left,t_train_right)['index']
            train_sample_numbers = len(train_indexes)

        if isinstance(test_size,(pd.Timestamp,list)):
            if isinstance(test_size,pd.Timestamp):
                t_test_left  = test_size
                t_test_right = None
            elif isinstance(test_size,list):
                check_list_assert(test_size)
                t_test_left = test_size[0]
                t_test_right = test_size[1]

            test_indexes  = df_vot.set_index('start').truncate(t_test_left, t_test_right)['index']
            test_sample_numbers = len(test_indexes)

    if isinstance(train_size,float):
        train_sample_numbers = int(len(X)*train_size)
    if isinstance(test_size,float):
        test_sample_numbers = int(len(X)*test_size)

    if isinstance(train_size,int):
        train_sample_numbers = train_size
    if isinstance(test_size,int):
        test_sample_numbers = test_size

    if train_sample_numbers is None: # due to test_size is not defined 
        train_sample_numbers = len(X) - test_sample_numbers
    if test_sample_numbers is None: # due to test_size is not defined 
        test_sample_numbers = len(X) - train_sample_numbers

    if train_sample_numbers + test_sample_numbers > len(X):
        raise Exception("There is not enough data in df dataset, according to the parameters train_size and test_size")



    ####### Part 2: Shuffle and generation of sets

    ind_train_left  = len(X)-1*(train_sample_numbers+test_sample_numbers) 
    ind_train_right = len(X)-test_sample_numbers
    ind_test_left  = len(X)-test_sample_numbers 
    ind_test_right = len(X) 

    assert ind_test_left>=ind_train_right
    assert (ind_test_right - ind_test_left) +  (ind_train_right - ind_train_left) <= len(df)



    if what_to_shuffle == 'nothing':
        X_train = X[ind_train_left:ind_train_right]
        y_train = y[ind_train_left:ind_train_right]
        X_test = X[ind_test_left:ind_test_right]
        y_test = y[ind_test_left:ind_test_right]

    elif what_to_shuffle == 'train':
        X_test = X[ind_test_left:ind_test_right]
        y_test = y[ind_test_left:ind_test_right] 

        X_train_meta = X[:ind_train_right]
        y_train_meta = y[:ind_train_right]

        indices = np.array(range(len(X_train_meta)))
        np.random.seed(random_state)
        np.random.shuffle(indices)
        indices=indices[:train_sample_numbers]
        X_train  = [X_train_meta[i] for i in indices]
        y_train  = [y_train_meta[i] for i in indices]

    elif what_to_shuffle == 'all':
        indices = np.array(range(len(X)))
        np.random.seed(random_state)
        np.random.shuffle(indices)
        X = [X[i] for i in indices]
        y = [y[i] for i in indices]

        X_train = X[ind_train_left:ind_train_right]
        y_train = y[ind_train_left:ind_train_right]
        X_test = X[ind_test_left:ind_test_right]
        y_test = y[ind_test_left:ind_test_right]
    else:
        raise Exception('Choose correct what_to_shuffle')
    return X_train, X_test, y_train, y_test
    
    
        
def split_by_repeated(series,df=None):
    """
    retrun dict with lists of ts whwre keys is unique values
    ts[ts.diff()!=0]  побыстрее будет
    """
    series = series.copy().dropna()
    if len(series.unique())==1:
        result = {series.unique()[0]:series}
    elif len(series.unique())>1:
        result = {uni:[] for uni in series.unique()}
        recent_i=0
        recent_val=series.values[0]
        for i in range(len(series)):
            val = series.values[i]
            if (recent_val == val):
                continue
            else:
                result[recent_val].append(series[recent_i:i])
                recent_i=i
                recent_val = val

        if i == len(series)-1:
            if (recent_val == val):
                result[recent_val].append(series[recent_i:i+1])
            else:
                result[recent_val].append(series[recent_i:i+1])
    else:
        raise NameError('0 series')


    if df is not None:
        new_result = {uni:[] for uni in series.unique()}
        for key in result:
            for i in range(len(result[key])):
                if len(result[key][i]) <=1:
                    continue
                else:
                    new_result[key].append(df.loc[result[key][i].index])
        return new_result
    else:
        return result
        
        
def df2dfs(df,  # Авторы не рекомендуют так делать,
            resample_freq = None, # требования
            thereshold_gap = None, 
            koef_freq_of_gap = 1.2, # 1.2 проблема которая возникает которую 02.09.2021 я написал в ИИ 
            plot = False,
            col = None):
    """
    Функция которая преообратает raw df до требований к входу на DL_AD    
    то есть разбивает df на лист of dfs by gaps 
    
    Не ресемлирует так как это тяжелая задача, но если частота реже чем 
    koef_freq_of_gap of thereshold_gap то это воспринимается как пропуск. 
    Основной посыл: если сигнал приходит чаще, то он не уползает сильно, 
    а значит не приводит к аномалии, а если редко то приводит, поэтому воспри-
    нимается как пропуск. 
    
    plot - очень долго
    
    Parameters
    ----------   
    df : pd.DataFrame
        Исходный временной ряд полностью за всю историю   
        
    resample_freq: pd.TimeDelta (optional, default=None)
        Частота дискретизации временного ряда. 
        Если default то самая частая частота дискретизации. При этом, 
        если нет выраженной частоты вылетит ошибка. 
    thereshold_gap : pd.TimeDelta (optional, default=None)
        Порог периода, превышая который функция будет воспринимать
        данный период как пропуск. 
    koef_freq_of_gap : float or int (optional if thereshold_gap==None,
        default=1.2)  
        thereshold_gap = koef_freq_of_gap * resample_freq
    plot : bool (optional, default=True)
        If true, then отрисуется нарезка
        If false, then не отрисуется нарезка   
    col : int of str (optional, default=True)
        Название или номер колонки для отрисовки
        Если None первая колонка
    Returns
    -------
    dfs : list of pd.DataFrame
        Список времменных рядов без пропусков с относительно стабильной 
        частотой дискретизации. 
    """

    df = df.dropna(how='all').dropna(1,how='all')
    dts  = df.dropna(how='all').index.to_series().diff()
    if resample_freq is None:
        dts_dist = dts.value_counts()
        if dts_dist[0] > dts_dist[1:].sum():
            resample_freq  = dts_dist.index[0]
        else: 
            print(dts_dist)
            raise Exception("Необходимо самостоятельно обработать функцию так как нет преобладающей частоты дискретизации")
    thereshold_gap = resample_freq*koef_freq_of_gap if thereshold_gap is None else thereshold_gap
    gaps = (dts > thereshold_gap).astype(int).cumsum()
    dfs = [df.loc[gaps[gaps==stage].index] for stage in gaps.unique()]
    
    if plot:
        f, ax = plt.subplots()
        if col is None:
            col = df.columns[0]
        else:
            if type(col)==type(int):
                col = df.columns[col]
        for df in dfs:
            df[col].plot(ax=ax)
    return dfs

    


def (df,num_points):
    """ 
    Посмотреть среднее расстояние между всеми парами точек (сэмлов) 
    в первых num_points точках
    в последних num_points точках
    и одновременно в первых и последжних num_points точках.
    """
    import itertools
    import scipy 
    array1 = df.iloc[:num_points].values
    array2 = df.iloc[num_points:int(2*num_points)].values

    indexes = list(range(len(array1)))
    pairs = list(set(itertools.permutations(indexes, 2)))
    list1 = array1[np.array(pairs)[:,0]]
    list2 = array1[np.array(pairs)[:,1]]
    print('Claster 1', scipy.spatial.distance.cdist(list1,list2).mean())


    indexes = list(range(len(array2)))
    pairs = list(set(itertools.permutations(indexes, 2)))
    list1 = array2[np.array(pairs)[:,0]]
    list2 = array2[np.array(pairs)[:,1]]
    print('Claster 2', scipy.spatial.distance.cdist(list1,list2).mean())

    common_array = np.concatenate([array1,array2])
    indexes = list(range(len(common_array)))
    pairs = list(set(itertools.permutations(indexes, 2)))
    list1 = common_array[np.array(pairs)[:,0]]
    list2 = common_array[np.array(pairs)[:,1]]
    print('Claster common', scipy.spatial.distance.cdist(list1,list2).mean())



from IPython import display
import torch
from ..useful.iterators import Loader
from matplotlib import pyplot as plt
import numpy as np
import random

#     criterion = weighted_mse_loss
#     optimizer = optim.Adam(net.parameters(), lr=lr)
#     model = model()

def set_determenistic(seed=None,precision=10):
    if seed is None:
        return
    np.random.seed(seed)
    random.seed(seed)
    #torch.backends.cudnn.benchmark = False
    #torch.backends.cudnn.deterministic = True
    torch.manual_seed(seed)
    #torch.set_printoptions(precision=precision)


def fit(
    model,
    optimiser,
    criterion,
    res_train_test_split,
    n_epochs = 10,
    batch_size = 2056,
    best_model_file =  './best_ae.pth', #None
    points_ahead = 1,
    random_state = None,
    show_progress = True,
    title=None,
    
       ):
    
    X_train, X_test, y_train, y_test = res_train_test_split[0],\
    res_train_test_split[1], res_train_test_split[2],res_train_test_split[3] 
    #     all_loader = Loader(scaled_research_df, scaled_research_df, batch_size, shuffle=False)
    train_iterator = Loader(X_train, y_train, batch_size, shuffle=True, random_state=random_state) # !!! кастыль
    val_iterator = Loader(X_test, y_test, batch_size, shuffle=False)
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        
    history_train = []
    history_val = []
    best_val_loss = float('+inf')
    show_progress_text =""
    for epoch in range(n_epochs):
        train_loss = model.run_epoch(train_iterator, optimiser, criterion, phase='train',
                                        points_ahead = points_ahead, device=device)  # , writer=writer)
        val_loss = model.run_epoch(val_iterator, None, criterion, phase='val',
                                        points_ahead = points_ahead, device=device)  # , writer=writer)

        history_train.append(train_loss)
        history_val.append(val_loss)

        if val_loss < best_val_loss:
            best_val_loss = val_loss
            if best_model_file is not None:
                torch.save(model.state_dict(), best_model_file)

        display.clear_output(wait=True)
        plt.figure()
        plt.plot(history_train, label=f'Train  {round(train_loss,3)}')
        plt.plot(history_val, label=f'Val  {round(val_loss,3)}')
        plt.xlabel('Epoch')
        plt.ylabel('MSE')
        plt.legend()
        if title is not None:
            plt.title(title)
        plt.grid()
        plt.show()
        if show_progress:
            show_progress_text = f'Epoch: {epoch + 1:02} \n' + \
                                 f'\tTrain Loss: {train_loss:.3f} \n' + \
                                 f'\t Val. Loss: {val_loss:.3f} \n\n' +  \
                                 show_progress_text
            print(show_progress_text)

    if best_model_file is not None:
        model.load_state_dict(torch.load(best_model_file))
    return model.run_epoch(val_iterator, None, criterion, phase='val', device=device) 